from BTrees.OOBTree import OOBTree
from Globals import InitializeClass
from Acquisition import aq_base, aq_chain
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from ComputedAttribute import ComputedAttribute

from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.utils import getToolByName

from Products.CMFPlone.CatalogTool import CatalogTool
from Products.CMFPlone.PloneFolder import BasePloneFolder

from Products.PlonePortlets.config import FILTEROUT
from Products.PlonePortlets.config import HIDE_PORTLET_ERRORS

from Products.PlonePortlets.permissions import VIEW_PERMISSION
from Products.PlonePortlets.permissions import MANAGE_PORTAL_PERMISSION
from Products.PlonePortlets.permissions import SET_OWN_PROPERTIES_PERMISSION
from Products.PlonePortlets.permissions import MODIFY_CONTENT_PERMISSION

from Products.PlonePortlets.interfaces import IPortletsTool


class PortletsTool(CatalogTool, BasePloneFolder, UniqueObject):
    """Tool for managing portlets."""

    id = 'portal_portlets'
    toolicon = 'skins/plone_images/book_icon.gif'

    meta_type = 'PortletsTool'
    archetype_name = 'PortletsTool'

    __implements__ = (IPortletsTool, CatalogTool.__implements__)

    security = ClassSecurityInfo()

    actions = (
        {
            'id': 'view',
            'name': 'View',
            'action': 'string:${object_url}/folder_contents',
            'permissions': (MODIFY_CONTENT_PERMISSION,),
        },
    )

    def _getPortalTypeName(self):
        """Returns the portal_type name."""
        return "PortletsTool"

    def manage_afterAdd(self, item, container):
        """XXX"""
        self.title = 'Handles portlets related functionality'
        self.memberPortlets = OOBTree()
        self.groupPortlets = {}
        self.portletsdata = {}

    security.declareProtected(VIEW_PERMISSION, 'getAvailablePortlets')
    def getAvailablePortlets(self):
        """Returns available portlets."""
        return self.searchResults()

    security.declareProtected(VIEW_PERMISSION, 'getHelperCSS')
    def getHelperCSS(self, context):
        """Helper CSS for all rendered portlets."""
        raise NotImplemented

    security.declareProtected(VIEW_PERMISSION, 'getHelperJS')
    def getHelperJS(self, context):
        """Helper JS for all rendered portlets."""
        raise NotImplemented

    security.declareProtected(VIEW_PERMISSION, 'findPortletsForContext')
    def findPortletsForContext(self, context):
        """Returns available portlet UIDS sorted."""
        #if hasattr(self,'_v_pdata'):
        #    return self._v_pdata
        structures = []
        # traverse the aquisition chain to find portlets added/filtered by context
        for item in aq_chain(context):
            if hasattr(aq_base(item), 'portletsdata'):
                structures.append(item.portletsdata.copy())
        # add the portets registered in the tool
        if self.portletsdata:
            structures.append(self.portletsdata)
        structures.reverse()

        # look up and add the portlets for groups the current user belongs to
        structures.extend(self.getMyGroupPortlets()[:])

        # finally the users own personal preferences.
        structures.append(self.getPersonalPortlets().copy())
        pdata = {}
        for struct in structures:
            filterOut = struct.get(FILTEROUT, [])[:]
            if filterOut:
                del struct[FILTEROUT]
            # filter out all the uids in filterOut
            for filteritem in filterOut:
                for key, categories in pdata.items():
                    if filteritem in categories:
                        cat = list(categories[:])
                        cat.remove(filteritem)
                        pdata[key] = cat

            for key,val in struct.items():
                l = pdata.get(key, [])
                if val:
                    if not val[0] in l:
                        l.extend(val)
                pdata[key] = l
        #self._v_pdata = pdata
        return pdata

    security.declareProtected(VIEW_PERMISSION, 'getPortletsSection')
    def getPortletsSection(self, context, section):
        """Gets all portlet objects in a category."""
        portletsToRender = self.findPortletsForContext(context).get(section,[])
        results = self.searchResults(UID=portletsToRender)[:]
        out = []
        # not the most efficient routine, but sorting portlets should work and this one can be replaced at any time.
        # feel free to replace
        # the results in r need to be output in the same order as their UIDs are listed in portletsToRender
        for uid in portletsToRender:
            for r in results:
                if r.UID == uid:
                    out.append(r)
        return out

    security.declareProtected(VIEW_PERMISSION, 'getPortletsSectionExplicit')
    def getPortletsSectionExplicit(self, context, section):
        """Gets all portlet objects in a category."""
        if hasattr(context.aq_base, 'portletsdata'):
            data = getattr(context.aq_base, 'portletsdata')
            uids = data.get(section, [])
            results = self.searchResults(UID=uids)
            return results
        else:
            return []

    def _rewrapPortlet(self, portlet, context):
        """Wraps a portlet in the current aquisition context, regardless of
        where it is located."""
        unwrapped = aq_base(portlet)
        rewrapped = unwrapped.__of__(context)
        return rewrapped

    security.declareProtected(VIEW_PERMISSION, 'wrapAndFilterSection')
    def wrapAndFilterSection(self, context, section):
        """Takes a section (left/right), acquisition-wraps it and returns it."""
        results = self.getPortletsSection(context, section)
        out = []
        for i in results:
            o = i.getObject()
            if o == context:
                continue
            rewrapped = self._rewrapPortlet(o, context)
            # This can make cache less efficient, if it executes queries anyway.
#            if rewrapped.portletVisibilityInContext(context):
#                out.append(rewrapped)
            out.append(rewrapped)
        return out

    security.declareProtected(VIEW_PERMISSION, 'renderPortletsSection')
    def renderPortletsSection(self, context, section):
        """Returns a string of html, rendering a group of portlets."""
        portlets = self.wrapAndFilterSection(context, section)
        out = ''
        for p in portlets:
            try:
                portletoutput = p()
                if portletoutput:
                    out += portletoutput
            except (KeyError, AttributeError),e:
                getToolByName(self, 'plone_utils').logException()
                if not HIDE_PORTLET_ERRORS:
                    raise e
        return out

    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'addPortletToSection')
    def addPortletToSection(self, context, portletUID, section):
        """Adds a portlet to a category/column."""
        data = {}
        if hasattr(aq_base(context), 'portletsdata'):
            data = context.portletsdata.copy()
            if data.has_key(section):
                l = list(data[section][:])
                l.append(portletUID)
                data[section] = tuple(l)
            else:
                data[section] = (portletUID,)
        else:
            data = {section:(portletUID,)}
        context.portletsdata = data

    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'removePortletFromSection')
    def removePortletFromSection(self, context, portletUID, section):
        """Removes a portlet from a category/column."""
        data = {}
        if hasattr(aq_base(context), 'portletsdata'):
            data = context.portletsdata.copy()
            if data.has_key(section):
                if portletUID in data[section]:
                    l = list(data[section][:])
                    l.remove(portletUID)
                    data[section] = tuple(l)
        context.portletsdata = data

    # XXX: Security?
    def addPortletToFilter(self, context, portletUID, section):
        """Adds a portlet to the ones filtered out in context."""
        pass

    # XXX: the personal settings - REMOVE-ME
    def addPersonalPortletToSection(self, portletUID, section):
        """XXX"""
        pd = self.getPersonalPortlets()
        gr = pd.get(section, [])[:]
        if not portletUID in gr:
            gr.append(portletUID)
        pd[section] = gr
        self.setPersonalPortlets(pd)
        #not done

    security.declareProtected(SET_OWN_PROPERTIES_PERMISSION, 'setPersonalFilter')
    def setPersonalFilter(self, portletUIDs):
        """XXX"""
        data = self.getPersonalPortlets()
        if not data.has_key(FILTEROUT):
            data[FILTEROUT] = []
        data[FILTEROUT] = tuple(portletUIDs)
        self.setPersonalPortlets(data)

    security.declareProtected(SET_OWN_PROPERTIES_PERMISSION, 'setPersonalPortlets')
    def setPersonalPortlets(self, data):
        """XXX"""
        username = getSecurityManager().getUser().getUserName()
        self.memberPortlets[username] = data

    security.declareProtected(SET_OWN_PROPERTIES_PERMISSION, 'getPersonalPortlets')
    def getPersonalPortlets(self):
        """XXX"""
        username = getSecurityManager().getUser().getUserName()
        return self.memberPortlets.get(username, {}).copy()

    security.declareProtected(SET_OWN_PROPERTIES_PERMISSION, 'setPersonalPortletsBySection')
    def setPersonalPortletsBySection(self, portletUIDs, section):
        """XXX"""
        data = self.getPersonalPortlets()
        portlets = data.get(section, [])
        portlets = tuple(portletUIDs)
        data[section] = portlets
        self.setPersonalPortlets(data)
        # notify the persistence-machinery
        self._p_changed = 1

    security.declareProtected(VIEW_PERMISSION, 'getPersonalPortletsBySection')
    def getPersonalPortletsBySection(self, section):
        """XXX"""
        ps = self.getPersonalPortlets().get(section, [])
        results = self.searchResults(UID=ps)
        return results

    # GROUP METHODS

    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'getMyGroupPortlets')
    def getMyGroupPortlets(self):
        """XXX"""
        grouptool = getToolByName(self, 'portal_groups')
        username = getSecurityManager().getUser().getUserName()
        groupids = [i.getGroupId() for i in grouptool.getGroupsByUserId(username)]
        portletdefinitions = []
        for gr in groupids:
            data = self.getPortletsForUserGroup(gr)
            if data:
                portletdefinitions.append(data.copy())
        return portletdefinitions

    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'getPortletsForUserGroup')
    def getPortletsForUserGroup(self, usergroup):
        """XXX"""
        prefix = self.acl_users.getGroupPrefix()
        if prefix and not usergroup.startswith(prefix):
            usergroup = prefix + usergroup
        return self.groupPortlets.get(usergroup,{})

    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'getPortletsForUserGroupBySection')
    def getPortletsForUserGroupBySection(self, usergroup, section):
        """XXX"""
        data = self.getPortletsForUserGroup(usergroup)
        portletuids = data.get(section, [])
        if not portletuids:
            return []
        return self.searchResults(UID=portletuids)

    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'setPortletsForUserGroup')
    def setPortletsForUserGroup(self, usergroup, data):
        """XXX"""
        prefix = self.acl_users.getGroupPrefix()
        #self.groupPortlets[prefix+usergroup] = data
        groupportlets = self.groupPortlets.copy()
        groupportlets[prefix+usergroup] = data
        self.groupPortlets = groupportlets


    def setPortletFilters(self, portletUIDs, context=None, usergroup=None):
        """set filters to a context or a group"""
        self.setPortletsSection(portletUIDs, FILTEROUT, context=context, usergroup=usergroup)

    # the general setter for contexts or groups
    security.declareProtected(MANAGE_PORTAL_PERMISSION, 'setPortletsSection')
    def setPortletsSection(self, portletUIDs, section, context=None, usergroup=None):
        """XXX"""
        if usergroup:
            gd = self.getPortletsForUserGroup(usergroup)
            gd[section] = portletUIDs
            self.setPortletsForUserGroup(usergroup,gd)
        else:
            if context is None:
                context = self
            if hasattr(context.aq_base,'portletsdata'):
                portletsdata = context.portletsdata.copy()
            else:
                portletsdata = {}
            portletsdata[section] = portletUIDs
            context.portletsdata = portletsdata

    def _verifyObjectPaste(self, object, validate_src=1):
        """XXX"""
        return BasePloneFolder._verifyObjectPaste(self, object, validate_src=1)


InitializeClass(PortletsTool)
